/*

   deque.c 

   Copyright, 1993, Brent Benson.  All Rights Reserved.

   Permission to use, copy, and modify this software and its
   documentation is hereby granted only under the following terms and
   conditions.  Both the above copyright notice and this permission
   notice must appear in all copies of the software, derivative works
   or modified version, and both notices must appear in supporting
   documentation.  Users of this software agree to the terms and
   conditions set forth in this notice and agree to inform the author,
   Brent Benson, of any noteworthy uses of this software.

*/

#include "deque.h"
#include "prim.h"
#include "symbol.h"
#include "error.h"

/* primitives */

static Object push (Object d, Object new);
static Object pop (Object d);
static Object push_last (Object d, Object new);
static Object pop_last (Object d);
static Object deque_first (Object d);
static Object deque_last (Object d);
static Object deque_element (Object d, Object i);
static Object deque_element_setter (Object d, Object i, Object new);
static Object deque_initial_state (Object d);
static Object deque_next_state (Object d, Object s);
static Object deque_final_state (Object d);
static Object deque_previous_state (Object d, Object s);
static Object deque_current_element (Object d, Object s);

static struct primitive deque_prims[] =
{
  {"%push", prim_2, push},
  {"%pop", prim_1, pop},
  {"%push-last", prim_2, push_last},
  {"%pop-last", prim_1, pop_last},
  {"%deque-first", prim_1, deque_first},
  {"%deque-last", prim_1, deque_last},
  {"%deque-element", prim_2, deque_element},
  {"%deque-element-setter", prim_3, deque_element_setter},
  {"%deque-initial-state", prim_1, deque_initial_state},
  {"%deque-next-state", prim_2, deque_next_state},
  {"%deque-final-state", prim_1, deque_final_state},
  {"%deque-previous-state", prim_2, deque_previous_state},
  {"%deque-current-element", prim_2, deque_current_element},
};

void 
init_deque_prims (void)
{
  int num;

  num = sizeof (deque_prims) / sizeof (struct primitive);
  init_prims (num, deque_prims);
}

Object
make_deque (void)
{
  Object obj;

  obj = allocate_object (sizeof (struct deque));
  DEQUETYPE (obj) = Deque;
  DEQUEFIRST (obj) = empty_list;
  DEQUELAST (obj) = empty_list;
  return (obj);
}

Object
make_deque_entry (Object prev, Object value, Object next)
{
  Object obj;

  obj = allocate_object (sizeof (struct deque_entry));
  DETYPE (obj) = DequeEntry;
  DEPREV (obj) = prev;
  DEVALUE (obj) = value;
  DENEXT (obj) = next;
  return (obj);
}

Object
make_deque_driver (Object args)
{
  int size;
  Object size_obj, fill_obj, entry, first, last, deq;
  Object size_keyword, fill_keyword;

  size = 0;
  size_obj = NULL;
  fill_obj = NULL;
  size_keyword = make_keyword ("size:");
  fill_keyword = make_keyword ("fill:");
  while (! NULLP (args))
    {
      if (FIRST (args) == size_keyword)
	{
	  size_obj = SECOND (args);
	}
      else if (FIRST (args) == fill_keyword)
	{
	  fill_obj = SECOND (args);
	}
      else
	{
	  error ("make: unsupported keyword for <deque> class", FIRST (args), NULL);
	}
      args = CDR (CDR (args));
    }
  if (size_obj)
    {
      if (! INTEGERP (size_obj))
	{
	  error ("make: value of size: argument must be an integer", size_obj, NULL);
	}
      size = INTVAL (size_obj);
    }
  if (! fill_obj)
    {
      fill_obj = false_object;
    }
  
  deq = make_deque ();
  /* actually fabricate the list representing the deque */
  if (size == 0)
    {
      DEQUEFIRST(deq) = DEQUELAST(deq) = empty_list;
      return (deq);
    }
  else
    {
      first = NULL;
      entry = empty_list;
      last = NULL;
      while (size)
	{
	  entry = make_deque_entry (empty_list, fill_obj, last);
	  if ( last )
	    {
	      DEPREV (last) = entry;
	    }
	  last = entry;
	  if ( ! first )
	    {
	      first = entry;
	    }
	  size--; 
	}
      DEQUEFIRST (deq) = first;
      DEQUELAST (deq) = last;
      return (deq);
    }
}
   
/* primitives */

static Object 
push (Object d, Object new)
{
  Object new_entry;

  new_entry = make_deque_entry (empty_list, new, DEQUEFIRST (d));
  if (DEQUEFIRST(d) == empty_list)
    {
      DEQUEFIRST(d) = DEQUELAST(d) = new_entry;
    }
  else
    {
      DEPREV (DEQUEFIRST (d)) = new_entry;
      DEQUEFIRST (d) = new_entry;
    }
  return (d);
}

static Object 
pop (Object d)
{
  Object ret;

  if (DEQUEFIRST(d) == empty_list)
    {
      error ("pop: cannot pop empty <deque>", d, NULL);
    }
  ret = DEVALUE (DEQUEFIRST (d));
  DEQUEFIRST (d) = DENEXT (DEQUEFIRST (d));
  if (DEQUEFIRST(d) != empty_list)
    {
      DEPREV (DEQUEFIRST (d)) = empty_list;
    }
  return (ret);
}

static Object 
push_last (Object d, Object new)
{
  Object new_entry;

  new_entry = make_deque_entry (DEQUELAST(d), new, empty_list);
  if (DEQUEFIRST(d) == empty_list)
    {
      DEQUEFIRST(d) = DEQUELAST(d) = new_entry;
    } 
  else
    {
      DENEXT (DEQUELAST (d)) = new_entry;
      DEPREV (new_entry) = DEQUELAST (d);
      DEQUELAST (d) = new_entry;
    }
  return (d);
}

static Object 
pop_last (Object d)
{
  Object res;

  if (DEQUEFIRST(d) == empty_list)
    {
      error ("pop-list: cannot pop empty <deque>", d, NULL);
    }
  res = DEVALUE (DEQUELAST (d));
  if (DEQUEFIRST(d) == DEQUELAST(d))
    {
      DEQUEFIRST(d) = DEQUELAST(d) = empty_list;
    }
  else
    {
      DEQUELAST(d) = DEPREV (DEQUELAST(d));
      if (DEQUELAST(d) != empty_list)
	{
	  DENEXT (DEQUELAST(d)) = empty_list;
	}
    }
  return (res);
}

static Object 
deque_first (Object d)
{
  if (DEQUEFIRST(d) == empty_list)
    {
      error ("first: empty <deque>", d, NULL);
    }
  return (DEVALUE (DEQUEFIRST (d)));
}

static Object 
deque_last (Object d)
{
  if (DEQUELAST(d) == empty_list)
    {
      error ("last: empty <deque>", d, NULL);
    }
  return (DEVALUE (DEQUELAST (d)));
}

static Object 
deque_element (Object d, Object index)
{
  int i;
  Object el;

  i = INTVAL (index);
  el = DEQUEFIRST (d);
  while ( i )
    {
      i--;
      el = DENEXT (el);
      if (el == empty_list)
	{
	  error ("element: out of range", index, d, NULL);
	}
    }
  return (DEVALUE (el));
}

static Object 
deque_element_setter (Object d, Object index, Object new)
{
  int i;
  Object el;

  i = INTVAL (index);
  el = DEQUEFIRST (d);
  while ( i )
    {
      i--;
      el = DENEXT (el);
      if (el == empty_list)
	{
	  error ("element: out of range", index, d, NULL);
	}
    }
  DEVALUE (el) = new;
  return (unspecified_object);
}

static Object 
deque_initial_state (Object d)
{
  if (DEQUEFIRST(d) == empty_list)
    {
      return (false_object);
    }
  else
    {
      return (DEQUEFIRST(d));
    }
}

static Object 
deque_next_state (Object d, Object s)
{
  if (DENEXT(s) == empty_list)
    {
      return (false_object);
    }
  else
    {
      return (DENEXT (s));
    }
}

static Object 
deque_final_state (Object d)
{
  if (DEQUELAST(d) == empty_list)
    {
      return (false_object);
    }
  else
    {
      return (DEQUELAST(d));
    }
}

static Object 
deque_previous_state (Object d, Object s)
{
  if (DEPREV(s) == empty_list)
    {
      return (false_object);
    }
  else
    {
      return (DEPREV (s));
    }
}

static Object 
deque_current_element (Object d, Object s)
{
  return (DEVALUE (s));
}

