/*

   table.c

   Copyright, 1993, Brent Benson.  All Rights Reserved.

   Permission to use, copy, and modify this software and its
   documentation is hereby granted only under the following terms and
   conditions.  Both the above copyright notice and this permission
   notice must appear in all copies of the software, derivative works
   or modified version, and both notices must appear in supporting
   documentation.  Users of this software agree to the terms and
   conditions set forth in this notice and agree to inform the author,
   Brent Benson, of any noteworthy uses of this software.

*/

#include "table.h"
#include "alloc.h"
#include "number.h"
#include "apply.h"
#include "list.h"
#include "symbol.h"
#include "error.h"

/* local function prototypes */

static Object table_element (Object table, Object key, Object def);
static Object table_element_setter (Object table, Object key, Object val);
static Object table_initial_state (Object table);
static Object table_next_state (Object table, Object state);
static Object table_current_element (Object table, Object state);
static Object table_current_key (Object table, Object state);
static Object table_current_element_setter (Object table, Object state, Object value);
static Object equal_hash (Object key);
static Object hash_pair (Object pair);
static Object hash_string (Object string);
static Object hash_vector (Object vector);

/* primitives */

static struct primitive table_prims[] =
{
  {"%table-element", prim_3, table_element},
  {"%table-element-setter", prim_3, table_element_setter},
  {"%table-initial-state", prim_1, table_initial_state},
  {"%table-next-state", prim_2, table_next_state},
  {"%table-current-element", prim_2, table_current_element},
  {"%table-current-key", prim_2, table_current_key},
  {"%table-current-element-setter", prim_3, table_current_element_setter},
  {"%=hash", prim_1, equal_hash},
};

void 
init_table_prims (void)
{
  int num;

  num = sizeof (table_prims) / sizeof (struct primitive);
  init_prims (num, table_prims);
}

Object 
make_table (int size)
{
  Object obj;

  obj = allocate_object (sizeof (struct table));
  TABLETYPE (obj) = Table;
  TABLESIZE (obj) = size;
  TABLETABLE (obj) = (Object *) checking_malloc (sizeof (Object) * size);
  memset (TABLETABLE (obj), 0, sizeof(Object) * size);
  return (obj);
}

Object
make_table_entry (int row, Object key, Object value, Object next)
{
  Object obj;

  obj = allocate_object (sizeof (struct table_entry));
  TETYPE (obj) = TableEntry;
  TEROW (obj) = row;
  TEKEY (obj) = key;
  TEVALUE (obj) = value;
  TENEXT (obj) = next;
  return (obj);
}

Object 
make_table_driver (Object rest)
{
  Object size_keyword, size;
  
  size_keyword = make_keyword ("size:");
  if (NULLP (rest))
    {
      return (make_table (DEFAULT_TABLE_SIZE));
    }
  else if (CAR (rest) == size_keyword)
    {
      rest = CDR (rest);
      if (NULLP (rest))
	{
	  error ("make: no argument given to size keyword", NULL);
	}
      size = CAR (rest);
      if (! INTEGERP (size))
	{
	  error ("make: argument to size keyword must be an integer", size, NULL);
	}
      return (make_table (INTVAL (size)));
    }
  else
    {
      error ("make: bad keywords or arguments", rest, NULL);
    }
}

/* local functions */

static Object 
table_element (Object table, Object key, Object def)
{
  Object hval, binary_equal_fun, entry;
  int h, old;

  hval = equal_hash (key);
  h = INTVAL (hval) % TABLESIZE (table);
  entry = TABLETABLE(table)[h];
  binary_equal_fun = symbol_value (make_symbol ("binary="));

  while ( entry )
    {
      if (apply (binary_equal_fun, cons (TEKEY (entry), cons (key, empty_list))) != false_object)
	{
	  return (TEVALUE (entry));
	}
      entry = TENEXT (entry);
    }
  if (def != unspecified_object)
    {
      return (def);
    }
  else
    {
      error ("element: no element matching key", key, NULL);
    }
}

static Object 
table_element_setter (Object table, Object key, Object val)
{
  Object hval, entry;
  int h;

  hval = equal_hash (key);
  h = INTVAL (hval) % TABLESIZE (table);
  entry = make_table_entry (h, key, val, TABLETABLE(table)[h]);
  TABLETABLE(table)[h] = entry;

  return (unspecified_object);
}

/* iteration protocol */

static Object 
table_initial_state (Object table)
{
  int i;

  for ( i=0 ; i < TABLESIZE (table) ; ++i )
    {
      if (TABLETABLE(table)[i])
	{
	  return (TABLETABLE(table)[i]);
	}
    }
  return (false_object);
}

static Object 
table_next_state (Object table, Object state)
{
  int i;

  if (TENEXT(state))
    {
      return (TENEXT(state));
    }
  for ( i=(TEROW(state)+1) ; i < TABLESIZE(table) ; ++i )
    {
      if (TABLETABLE(table)[i])
	{
	  return (TABLETABLE(table)[i]);
	}
    }
  return (false_object);
}

static Object 
table_current_element (Object table, Object state)
{
  return (TEVALUE(state));
}

static Object 
table_current_key (Object table, Object state)
{
  return (TEKEY(state));
}

static Object 
table_current_element_setter (Object table, Object state, Object value)
{
  TEVALUE (state) = value;
  return (unspecified_object);
}

static Object
equal_hash (Object key)
{
  Object hashfun;

  if (INSTANCEP (key))
    {
      hashfun = symbol_value (equal_hash_symbol);
      if (! hashfun)
	{
	  error ("no =hash method defined for key class", key, NULL);
	}
      return (apply (hashfun, cons (key, empty_list)));
    }
  else
    {
      if (INTEGERP (key))
	{
	  return (key);
	}
      else if (CHARP (key))
	{
	  return (make_integer (CHARVAL (key)));
	}
      else if (TRUEP (key))
	{
	  return (make_integer (1));
	}
      else if (FALSEP (key))
	{
	  return (make_integer (0));
	}
      else if (NULLP (key))
	{
	  return (make_integer (2));
	}
      else if (PAIRP (key))
	{
	  return (hash_pair (key));
	}
      else if (BYTESTRP (key))
	{
	  return (hash_string (key));
	}
      else if (SOVP (key))
	{
	  return (hash_vector (key));
	}
      else if (SYMBOLP (key) || KEYWORDP (key))
	{
	  return (make_integer ((int)key));
	}
      else
	{
	  error ("=hash: don't know how to hash object", key, NULL);
	}
    }
}

static Object 
hash_pair (Object pair)
{
  int h;

  h = INTVAL (equal_hash (CAR (pair))) + INTVAL (equal_hash (CDR (pair)));
  return (make_integer (h));
}

static Object 
hash_string (Object string)
{
  int i, h;

  h = 0;
  for ( i=0 ; i < BYTESTRSIZE (string) ; ++i )
    {
      h += BYTESTRVAL(string)[i];
    }
  return (make_integer (h));
}

static Object 
hash_vector (Object vector)
{
  int i, h;

  h = 0;
  for ( i=0 ; i < SOVSIZE (vector) ; ++i )
    {
      h += INTVAL (equal_hash (SOVELS(vector)[i]));
    }
  return (make_integer (h));
}
